// Windows and Windows DLLs
#include <windows.h>
#include <Xinput.h>
#include <dsound.h>

// Local imports
#include "handmade.h"
#include "win32_handmade.h"

global ATOM HH_CTRLW;
global bool globalRunning;
global Win32OffscreenBuffer globalBackBuffer;
global LPDIRECTSOUNDBUFFER globalSecondaryBuffer;
global int64 globalPerfCountFrequency;

// XInputGetState
#define X_INPUT_GET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_STATE *pState)
typedef X_INPUT_GET_STATE(XInputGetStateFn);
X_INPUT_GET_STATE(XInputGetStateStub) { return ERROR_DEVICE_NOT_CONNECTED; }
global XInputGetStateFn *XInputGetStateDyn = XInputGetStateStub;
#define XInputGetState XInputGetStateDyn

// XInputSetState
#define X_INPUT_SET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_VIBRATION *pVibration)
typedef X_INPUT_SET_STATE(XInputSetStateFn);
X_INPUT_SET_STATE(XInputSetStateStub) { return ERROR_DEVICE_NOT_CONNECTED; }
global XInputSetStateFn *XInputSetStateDyn = XInputSetStateStub;
#define XInputSetState XInputSetStateDyn

typedef HRESULT WINAPI DirectSoundCreateFn(LPCGUID pcGuidDevice, LPDIRECTSOUND *ppDS, LPUNKNOWN pUnkOuter);

#define stackAlloc(size, type) (type*)_alloca(size*sizeof(type))

DEBUG_PLATFORM_PRINTF(debug_printf) { 
    size_t bufsize = wcslen(format)*10;
    wchar_t *output = (wchar_t*)malloc(bufsize);
    va_list list;
    va_start(list, format);
    vswprintf(output, bufsize, format, list);
    OutputDebugStringW(output);
    free(output);
}

DEBUG_PLATFORM_FREE_FILE_MEMORY(debugFreeFileMemory) {
    if (fileMemory) {
        VirtualFree(fileMemory, NULL, MEM_RELEASE);
    }
}

DEBUG_PLATFORM_READ_ENTIRE_FILE(debugReadEntireFile) {
    DebugReadFileResult result = {};
    HANDLE fileHandle = CreateFileA(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, NULL, NULL);
    if (fileHandle != INVALID_HANDLE_VALUE) {
        LARGE_INTEGER fileSize;
        if (GetFileSizeEx(fileHandle, &fileSize)) {
            uint32 fileSize32 = safeTruncateUInt64(fileSize.QuadPart);
            result.contents = VirtualAlloc(NULL, fileSize.QuadPart, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
            if (result.contents) {
                DWORD bytesRead;
                if (ReadFile(fileHandle, result.contents, (DWORD)fileSize.QuadPart, &bytesRead, NULL) && (fileSize32 == (uint32)bytesRead)) {
                    result.contentsSize = fileSize32;
                } else {
                    debugFreeFileMemory(ctx, result.contents);
                    result.contents = NULL;
                    // logging
                }
            } else {
                result.contents = NULL;
                // logging
            }
        } else {
            // logging
        }
        CloseHandle(fileHandle);
    } else {
        // logging
    }
    return result;
}

DEBUG_PLATFORM_WRITE_ENTIRE_FILE(debugWriteEntireFile) {
    bool32 result = false;
    HANDLE fileHandle = CreateFileA(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, NULL, NULL);
    if (fileHandle != INVALID_HANDLE_VALUE) {
        DWORD bytesWritten;
        if (WriteFile(fileHandle, memory, memorySize, &bytesWritten, NULL)) {
            // file written successfully
            result = bytesWritten == memorySize;
        } else {
            // logging
        }
        CloseHandle(fileHandle);
    } else {
        // logging
    }
    return result;
}

internal void resizeDIBSection(Win32OffscreenBuffer *buffer, int width, int height) {
    if (buffer->memory) {
        VirtualFree(buffer->memory, NULL, MEM_RELEASE);
    }

    buffer->width = width;
    buffer->height = height;
    buffer->bytesPerPixel = 4;
    buffer->pitch = buffer->width*buffer->bytesPerPixel;

    buffer->info.bmiHeader.biSize = sizeof(buffer->info.bmiHeader);
    buffer->info.bmiHeader.biWidth = buffer->width;
    buffer->info.bmiHeader.biHeight = -buffer->height;
    buffer->info.bmiHeader.biPlanes = 1;
    buffer->info.bmiHeader.biBitCount = 32;
    buffer->info.bmiHeader.biCompression = BI_RGB;

    int bitmapSize = buffer->width*buffer->height*4;
    buffer->memory = VirtualAlloc(NULL, bitmapSize, MEM_COMMIT, PAGE_READWRITE);
}

internal Win32WindowDimensions win32GetWindowDimensions(HWND window) {
    Win32WindowDimensions result;
    RECT clientRect;
    GetClientRect(window, &clientRect);
    result.width = clientRect.right - clientRect.left;
    result.height = clientRect.bottom - clientRect.top;
    return result;
}

internal void win32DrawBufferInWindow(Win32OffscreenBuffer *buffer, HWND window) {
    Win32WindowDimensions winDims = win32GetWindowDimensions(window);
    HDC dc = GetDC(window);

    int offsetX = 10;
    int offsetY = 10;

    PatBlt(dc, 0, 0, winDims.width, offsetY, BLACKNESS);
    PatBlt(dc, 0, offsetY + buffer->height, winDims.width, winDims.height, BLACKNESS);
    PatBlt(dc, 0, 0, offsetX, winDims.height, BLACKNESS);
    PatBlt(dc, offsetX + buffer->width, 0, winDims.width, winDims.height, BLACKNESS);

    StretchDIBits(dc,
        offsetX, offsetY, buffer->width, buffer->height,
        0, 0, buffer->width, buffer->height,
        buffer->memory,
        &buffer->info, 
        DIB_RGB_COLORS,
        SRCCOPY
    );
}

inline FILETIME win32GetLastWriteTime(char *filename) {
    FILETIME lastWriteTime = {};
    WIN32_FIND_DATA findData;
    HANDLE findHandle = FindFirstFileA(filename, &findData);
    if (findHandle != INVALID_HANDLE_VALUE) {
        lastWriteTime = findData.ftLastWriteTime;
        FindClose(findHandle);
    }
    return lastWriteTime;
}

internal Win32GameCode win32LoadGameCode(char *filename, char *tempname) {
    Win32GameCode result = {};

    result.lastWriteTime = win32GetLastWriteTime(filename);

    CopyFile(filename, tempname, FALSE);
    HMODULE gameCodeLib = LoadLibrary(tempname);
    result.gameCodeLib = gameCodeLib;

    if (gameCodeLib) {
        result.updateAndRender = (GameUpdateAndRenderFn *)GetProcAddress(result.gameCodeLib, "gameUpdateAndRender");
        result.getSoundSamples = (GameGetSoundSamplesFn *)GetProcAddress(result.gameCodeLib, "gameGetSoundSamples");
        result.isValid = (result.updateAndRender && result.getSoundSamples);
    }
    if (!result.isValid) {
        result.getSoundSamples = gameGetSoundSamplesStub;
        result.updateAndRender = gameUpdateAndRenderStub;
    }

    return result;
}

internal void win32UnloadGameCode(Win32GameCode *gameCode) {
    if (gameCode->gameCodeLib) {
        FreeLibrary(gameCode->gameCodeLib);
    }
    gameCode->isValid = false;
    gameCode->getSoundSamples = gameGetSoundSamplesStub;
    gameCode->updateAndRender = gameUpdateAndRenderStub;
}

internal void win32LoadXInput() {
    HMODULE xInputLib = LoadLibrary("xinput1_4.dll");
    if (!xInputLib) {
        xInputLib = LoadLibrary("xinput1_3.dll");
    }

    if (xInputLib) {
        XInputGetState = (XInputGetStateFn *)GetProcAddress(xInputLib, "XInputGetState");
        XInputSetState = (XInputSetStateFn *)GetProcAddress(xInputLib, "XInputSetState");
    }
}

internal void win32ProcessKeyboardKeypress(GameButtonState *newState, bool32 isDown) {
    if (newState->endedDown != isDown) {
        newState->endedDown = isDown;
        newState->halfTransitionCount++;
    }
}

internal void win32ProcessXInputDigitalButton(DWORD xInputButtonState, GameButtonState *oldState, GameButtonState *newState, DWORD buttonBit) {
    newState->endedDown = (xInputButtonState & buttonBit) == buttonBit;
    newState->halfTransitionCount = (oldState->endedDown != newState->endedDown) ? 1 : 0;
}

internal void win32InitSound(HWND window, int32 samplesPerSec, int bufferSize) {
    HMODULE dSoundLibrary = LoadLibraryA("dsound.dll");
    if (dSoundLibrary) {
        DirectSoundCreateFn *DirectSoundCreate = (DirectSoundCreateFn*)GetProcAddress(dSoundLibrary, "DirectSoundCreate");
        LPDIRECTSOUND DirectSound;
        if (DirectSoundCreate && SUCCEEDED(DirectSoundCreate(0, &DirectSound, 0))) {
            WAVEFORMATEX waveFormat = {};
            waveFormat.wFormatTag = WAVE_FORMAT_PCM;
            waveFormat.nChannels = 2;
            waveFormat.nSamplesPerSec = samplesPerSec;
            waveFormat.wBitsPerSample = 16;
            waveFormat.nBlockAlign = (waveFormat.nChannels*waveFormat.wBitsPerSample) / 8;
            waveFormat.nAvgBytesPerSec = waveFormat.nSamplesPerSec*waveFormat.nBlockAlign;
            waveFormat.cbSize = 0;

            if (SUCCEEDED(DirectSound->SetCooperativeLevel(window, DSSCL_PRIORITY))) {
                DSBUFFERDESC bufferDescription = {};
                bufferDescription.dwSize = sizeof(bufferDescription);
                bufferDescription.dwFlags = DSBCAPS_PRIMARYBUFFER;

                LPDIRECTSOUNDBUFFER primaryBuffer;
                if (SUCCEEDED(DirectSound->CreateSoundBuffer(&bufferDescription, &primaryBuffer, 0))) {
                    if (SUCCEEDED(primaryBuffer->SetFormat(&waveFormat))) {
                        OutputDebugStringA("Primary buffer format was set.\n");
                    } else {
                        //No sound! (?)
                    }
                }
            } else {
                // No sound! (?)
            }

            DSBUFFERDESC bufferDescription = {};
            bufferDescription.dwSize = sizeof(bufferDescription);
            bufferDescription.dwFlags = 0;
            bufferDescription.dwBufferBytes = bufferSize;
            bufferDescription.lpwfxFormat = &waveFormat;

            if (SUCCEEDED(DirectSound->CreateSoundBuffer(&bufferDescription, &globalSecondaryBuffer, 0))) {
                OutputDebugStringA("Secondary buffer created successfully!\n");
            }
        } else {
            // No sound! (?)
        }
    }
}

LRESULT mainWindowCallback(HWND window, UINT message, WPARAM wParam, LPARAM lParam) {
    LRESULT result = 0;
    switch (message) {
        case WM_SIZE: {
        } break;

        case WM_DESTROY: {
            OutputDebugStringA("WM_DESTROY\n");
        } break;

        case WM_CLOSE: {
            globalRunning = false;
        } break;

        case WM_PAINT: {
            PAINTSTRUCT paint = {};
            HDC deviceContext = BeginPaint(window, &paint);
            int x = paint.rcPaint.left;
            int y = paint.rcPaint.top;
            int width = paint.rcPaint.right - paint.rcPaint.left;
            int height = paint.rcPaint.bottom - paint.rcPaint.top;
            win32DrawBufferInWindow(&globalBackBuffer, window);
            EndPaint(window, &paint);
        } break;

        case WM_ACTIVATEAPP: {
        } break;

        case WM_HOTKEY: {
            if (wParam == HH_CTRLW) {
                globalRunning = false;
            }
        } break;

        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
        case WM_KEYDOWN: {
            Assert(!"Keyboard input came in through a non-dispatch message.");
        } break;

        default: {
            result = DefWindowProcA(window, message, wParam, lParam);
        } break;
    }
    return result;
}

internal void win32ClearBuffer(Win32SoundOutput *soundOutput) {
    VOID *region1;
    DWORD region1Size;
    VOID *region2;
    DWORD region2Size;
    HRESULT lock = globalSecondaryBuffer->Lock(0, soundOutput->secondaryBufferSize, &region1, &region1Size, &region2, &region2Size, 0);
    if (SUCCEEDED(lock)) {
        int16 *destSamples = (int16*)region1;
        for (DWORD byteIndex = 0; byteIndex < region1Size; byteIndex++) {
           // *destSamples++ = 0;
        }
        destSamples = (int16*)region2;
        for (DWORD byteIndex = 0; byteIndex < region2Size; byteIndex++) {
            //*destSamples++ = 0;
        }
        globalSecondaryBuffer->Unlock(region1, region1Size, region2, region2Size);
    } else {
        // no lock!
    }
}

internal void win32FillSoundBuffer(Win32SoundOutput *soundOutput, DWORD byteToLock, DWORD bytesToWrite, GameSoundOutputBuffer *sourceBuffer) {
    VOID *region1;
    DWORD region1Size;
    VOID *region2;
    DWORD region2Size;
    HRESULT lock = globalSecondaryBuffer->Lock(byteToLock, bytesToWrite, &region1, &region1Size, &region2, &region2Size, 0);
    if (SUCCEEDED(lock)) {
        DWORD region1SampleCount = region1Size/soundOutput->bytesPerSample;
        int16 *destSamples = (int16*)region1;
        int16 *sourceSamples = sourceBuffer->samples;
        for (DWORD sampleIndex = 0; sampleIndex < region1SampleCount; sampleIndex++) {
            *destSamples++ = *sourceSamples++;
            *destSamples++ = *sourceSamples++;
            soundOutput->runningSampleIndex++;
        }
        DWORD region2SampleCount = region2Size/soundOutput->bytesPerSample;
        destSamples = (int16*)region2;
        for (DWORD sampleIndex = 0; sampleIndex < region2SampleCount; sampleIndex++) {
            *destSamples++ = *sourceSamples++;
            *destSamples++ = *sourceSamples++;
            soundOutput->runningSampleIndex++;
        }
        globalSecondaryBuffer->Unlock(region1, region1Size, region2, region2Size);
    } else {
        // no lock!
    }
}

internal Win32ReplayBuffer* win32GetReplayBuffer(Win32State *state, int unsigned index) {
    Win32ReplayBuffer *result = &state->replayBuffers[index - 1];
    return result;
}

internal void win32BeginRecordingInput(Win32State *state, int inputRecordingIndex) {
    Win32ReplayBuffer *replayBuffer = win32GetReplayBuffer(state, inputRecordingIndex);
    if (replayBuffer->memoryBlock) {
        state->inputRecordingIndex = inputRecordingIndex;
        state->recordingHandle = replayBuffer->fileHandle;
        LARGE_INTEGER filePosition;
        filePosition.QuadPart = state->gameMemoryTotalSize;
        SetFilePointerEx(state->recordingHandle, filePosition, NULL, FILE_BEGIN);
        CopyMemory(replayBuffer->memoryBlock, state->gameMemoryBlock, state->gameMemoryTotalSize);
    }
}

internal void win32EndRecordingInput(Win32State *win32State) {
    win32State->inputRecordingIndex = 0;
}

internal void win32RecordInput(Win32State *win32State, GameInput *newInput) {
    DWORD bytesWritten;
    WriteFile(win32State->recordingHandle, newInput, sizeof(*newInput), &bytesWritten, NULL);
}

internal void win32BeginInputPlayback(Win32State *state, int inputPlayingIndex) {
    Win32ReplayBuffer *replayBuffer = win32GetReplayBuffer(state, inputPlayingIndex); 
    if (replayBuffer->memoryBlock) {
        state->inputPlayingIndex = inputPlayingIndex;
        state->playbackHandle = replayBuffer->fileHandle;
        LARGE_INTEGER filePosition;
        filePosition.QuadPart = state->gameMemoryTotalSize;
        SetFilePointerEx(state->playbackHandle, filePosition, NULL, FILE_BEGIN);
        CopyMemory(state->gameMemoryBlock, replayBuffer->memoryBlock, state->gameMemoryTotalSize);
    }
}

internal void win32EndInputPlayback(Win32State *win32State) {
    win32State->inputPlayingIndex = 0;
}

internal void win32PlaybackInput(Win32State *win32State, GameInput *newInput) {
    DWORD bytesRead;
    if (ReadFile(win32State->playbackHandle, newInput, sizeof(*newInput), &bytesRead, NULL)) {
        if (bytesRead == 0) {
            // Restart
            int playingIndex = win32State->inputPlayingIndex;
            win32EndInputPlayback(win32State);
            win32BeginInputPlayback(win32State, playingIndex);
        } else {
            // there's still input, read next input
        }
    }
}

internal void win32ProcessPendingMessages(Win32State *win32State, GameControllerInput *keyboardController) {
    MSG message;
    while (PeekMessageA(&message, NULL, NULL, NULL, PM_REMOVE)) {
        if (message.message == WM_QUIT) {
            globalRunning = false;
        }

        switch (message.message) {
            case WM_SYSKEYDOWN:
            case WM_SYSKEYUP:
            case WM_KEYUP:
            case WM_KEYDOWN: {
                uint32 VKCode = (uint32)message.wParam;
                bool wasDown = (message.lParam & (1 << 30)) != 0;
                bool isDown = (message.lParam & (1 << 31)) == 0;
                if (wasDown != isDown) {
                    if (VKCode == 'W') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickUp, isDown);
                    } else if (VKCode == 'A') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickLeft, isDown);
                    } else if (VKCode == 'S') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickDown, isDown);
                    } else if (VKCode == 'D') {
                        win32ProcessKeyboardKeypress(&keyboardController->stickRight, isDown);
                    } else if (VKCode == 'Q') {
                        win32ProcessKeyboardKeypress(&keyboardController->leftShoulder, isDown);
                    } else if (VKCode == 'E') {
                        win32ProcessKeyboardKeypress(&keyboardController->rightShoulder, isDown);
                    } else if (VKCode == VK_ESCAPE) {
                    } else if (VKCode == VK_UP) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnUp, isDown);
                    } else if (VKCode == VK_LEFT) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnLeft, isDown);
                    } else if (VKCode == VK_DOWN) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnDown, isDown);
                    } else if (VKCode == VK_RIGHT) {
                        win32ProcessKeyboardKeypress(&keyboardController->btnRight, isDown);
                    } else if (VKCode == VK_SPACE) {
                    } else if (VKCode == 'L') {
                        if (isDown) {
                            if (win32State->inputRecordingIndex == 0 && win32State->inputPlayingIndex == 0) {
                                win32BeginRecordingInput(win32State, 1);
                            } else if (win32State->inputRecordingIndex != 0 && win32State->inputPlayingIndex == 0) {
                                win32EndRecordingInput(win32State);
                                win32BeginInputPlayback(win32State, 1);
                            } else if (win32State->inputRecordingIndex == 0 && win32State->inputPlayingIndex != 0) {
                                win32EndInputPlayback(win32State);
                            }
                        }
                    }
                }

                bool32 altKeyWasDown = message.lParam & (1 << 29);
                if (altKeyWasDown && VKCode == VK_F4) {
                    globalRunning = false;
                }
            } break;

            default: {
                TranslateMessage(&message);
                DispatchMessage(&message);
            } break;
        }
    }
}

inline LARGE_INTEGER win32GetWallClock() {
    LARGE_INTEGER result;
    QueryPerformanceCounter(&result);
    return result;
}

inline real32 win32GetSecondsElapsed(LARGE_INTEGER start, LARGE_INTEGER end) {
    return (real32)(end.QuadPart - start.QuadPart) / (real32)globalPerfCountFrequency;
}

internal void win32DebugDrawVertical(Win32OffscreenBuffer *buffer, int x, int top, int bottom, int32 color) {
    uint8 *pixel = (uint8 *)buffer->memory + top * buffer->pitch + x*buffer->bytesPerPixel;
    for (int y = top; y < bottom; y++) {
        *(uint32 *)pixel = color;
        pixel += buffer->pitch;
    }
}

inline void win32DrawSoundBufferMarker(Win32OffscreenBuffer *buffer, Win32SoundOutput *soundOutput, real32 pxPerSoundBufferEntry, int padX, int top, int bottom, DWORD value, uint32 color) {
    Assert(value < (DWORD)soundOutput->secondaryBufferSize);
    int x = padX + (int)(pxPerSoundBufferEntry * (real32)value);
    win32DebugDrawVertical(buffer, x, top, bottom, color);
}

void catStrings(size_t sourceACount, char *sourceA, size_t sourceBCount, char *sourceB, size_t destCount, char *dest) {
    for (int i = 0; i < sourceACount; i++) { 
        *dest++ = *sourceA++;
    }
    for (int i = 0; i < sourceBCount; i++) { 
        *dest++ = *sourceB++;
    }
    *dest++ = 0;
}

void win32GetExeFilename(Win32State *win32State) {
    DWORD sizeOfFilename = GetModuleFileNameA(NULL, win32State->exeFilename, sizeof(win32State->exeFilename));
    win32State->onePastLastExeFilenameSlash = win32State->exeFilename;
    for (char *scan = win32State->exeFilename; *scan; scan++) {
        if (*scan == '\\') {
            win32State->onePastLastExeFilenameSlash = scan + 1;
        }
    }
}

void win32GetFullPathToLocalFile(Win32State *state, char *filename, int filenameSize, char *maxPathBuffer) {
    catStrings(state->onePastLastExeFilenameSlash - state->exeFilename, state->exeFilename, filenameSize, filename, (size_t)MAX_PATH, maxPathBuffer);
}

#define WINDOW_WIDTH 960
#define WINDOW_HEIGHT 540

int APIENTRY WinMain(HINSTANCE instance, HINSTANCE prevInstance, PSTR commandLine, int commandShow) {
    Win32State win32State = {};
    win32GetExeFilename(&win32State);

    char *sourceGameCodeDLLFilename = "handmade.dll";
    char sourceGameCodeDLLFullPath[MAX_PATH];
    win32GetFullPathToLocalFile(&win32State, sourceGameCodeDLLFilename, sizeof(sourceGameCodeDLLFullPath) - 1, sourceGameCodeDLLFullPath);

    char tempGameCodeDLLFilename[] = "handmade_temp.dll";
    char tempGameCodeDLLFullPath[MAX_PATH];
    win32GetFullPathToLocalFile(&win32State, tempGameCodeDLLFilename, sizeof(tempGameCodeDLLFullPath) - 1, tempGameCodeDLLFullPath);

    LARGE_INTEGER performanceFrequencyResult;
    QueryPerformanceFrequency(&performanceFrequencyResult);
    globalPerfCountFrequency = performanceFrequencyResult.QuadPart;

    UINT desiredSchedulerMs = 1;
    bool32 sleepIsGranular = TIMERR_NOERROR == timeBeginPeriod(desiredSchedulerMs);

    WNDCLASSA windowClass = {};
    windowClass.style = CS_VREDRAW | CS_HREDRAW;
    windowClass.lpfnWndProc = &mainWindowCallback;
    windowClass.hInstance = instance;
    windowClass.lpszClassName = "HandmadeHeroWindowClass";
    resizeDIBSection(&globalBackBuffer, WINDOW_WIDTH, WINDOW_HEIGHT);

    if (RegisterClass(&windowClass)) {
        DWORD windowFlags = WS_OVERLAPPEDWINDOW | WS_VISIBLE;
        RECT rect = { 0, 0, WINDOW_WIDTH, WINDOW_HEIGHT };
        AdjustWindowRectEx(&rect, windowFlags, false, NULL);
        HWND window = CreateWindowExA(
            NULL,
            windowClass.lpszClassName,
            "Handmade Hero",
            windowFlags,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            rect.right - rect.left,
            rect.bottom - rect.top,
            NULL,
            NULL,
            instance,
            NULL
        );

        if (window) {
            int monitorRefreshHz = 60;
            int win32RefreshRate = GetDeviceCaps(GetDC(window), VREFRESH);
            if (win32RefreshRate > 1) {
                monitorRefreshHz = win32RefreshRate;
            }
            real32 gameUpdateHz = monitorRefreshHz / 2.0f;
            real32 targetSecondsPerFrame = 1.0f / (real32)gameUpdateHz;

            HH_CTRLW = GlobalAddAtomA("HH_CTRLW");
            RegisterHotKey(window, HH_CTRLW, MOD_CONTROL, 'W');

            globalRunning = true;

            GameInput input[2] = {};
            GameInput *oldInput = &input[0];
            GameInput *newInput = &input[1];

            win32LoadXInput();

            // sound test
            Win32SoundOutput soundOutput = {};
            soundOutput.samplesPerSecond = 48000;
            soundOutput.runningSampleIndex = 0;
            soundOutput.bytesPerSample = sizeof(int16)*2;
            soundOutput.secondaryBufferSize = soundOutput.samplesPerSecond*soundOutput.bytesPerSample;
            soundOutput.safetyBytes = (int)((soundOutput.samplesPerSecond * soundOutput.bytesPerSample / gameUpdateHz) / 2.0f);

            int16 *samples = (int16*)VirtualAlloc(NULL, soundOutput.secondaryBufferSize, MEM_COMMIT, PAGE_READWRITE);

#if HANDMADE_INTERNAL
            LPVOID baseAddress = (LPVOID)Terabytes((uint64)2);
#else
            LPVOID baseAddress = 0;
#endif
            GameMemory gameMemory = {};
            gameMemory.permanentStorageSize = Megabytes(64);
            gameMemory.transientStorageSize = Gigabytes((uint64)1);
            gameMemory.debugFreeFileMemory = debugFreeFileMemory;
            gameMemory.debugWriteEntireFile = debugWriteEntireFile;
            gameMemory.debugReadEntireFile = debugReadEntireFile;
            gameMemory.debug_printf = debug_printf;

            win32State.gameMemoryTotalSize = gameMemory.permanentStorageSize + gameMemory.transientStorageSize;
            win32State.gameMemoryBlock = VirtualAlloc(baseAddress, win32State.gameMemoryTotalSize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
            gameMemory.permanentStorage = win32State.gameMemoryBlock;
            gameMemory.transientStorage = ((uint8 *)gameMemory.permanentStorage + gameMemory.permanentStorageSize);

            for (int i = 0; i < ArrayCount(win32State.replayBuffers); i++) {
                // Filename
                char prefix[] = "replay-";
                char suffix[] = ".hmr";
                const int totalLen = ArrayCount(prefix) + 1 + ArrayCount(suffix);
                char replayFilename[totalLen] = {};
                char digit[] = {(char)(48 + i)};
                catStrings(ArrayCount(prefix) - 1, prefix, 1, digit, ArrayCount(replayFilename) - 1, replayFilename);
                catStrings(ArrayCount(prefix), replayFilename, ArrayCount(suffix) - 1, suffix, ArrayCount(replayFilename) - 1, replayFilename);
                win32GetFullPathToLocalFile(&win32State, replayFilename, sizeof(replayFilename) - 1, win32State.replayBuffers[i].replayFilename);
                Win32ReplayBuffer *replayBuffer = &win32State.replayBuffers[i];

                replayBuffer->fileHandle = CreateFileA(replayBuffer->replayFilename, GENERIC_WRITE | GENERIC_READ, NULL, NULL, CREATE_ALWAYS, NULL, NULL);
                DWORD maxSizeLow = (win32State.gameMemoryTotalSize & 0xFFFFFFFF);
                DWORD maxSizeHigh = (win32State.gameMemoryTotalSize >> 32);
                replayBuffer->memoryMap = CreateFileMappingA(replayBuffer->fileHandle, NULL, PAGE_READWRITE, maxSizeHigh, maxSizeLow, NULL);
                replayBuffer->memoryBlock = MapViewOfFile(replayBuffer->memoryMap, FILE_MAP_ALL_ACCESS, NULL, NULL, win32State.gameMemoryTotalSize);
            }

            win32InitSound(window, soundOutput.samplesPerSecond, soundOutput.secondaryBufferSize);
            win32ClearBuffer(&soundOutput);
            globalSecondaryBuffer->Play(0, 0, DSBPLAY_LOOPING);

            LARGE_INTEGER lastWorkCounter = win32GetWallClock();
            LARGE_INTEGER flipWallClock = win32GetWallClock();

            DWORD audioLatencyBytes = 0;
            real32 audioLatencySeconds = 0;
            bool soundIsValid = false;

            Win32GameCode game = win32LoadGameCode(sourceGameCodeDLLFullPath, tempGameCodeDLLFullPath);

            int64 lastCycleCount = __rdtsc();
            while (globalRunning) {
                newInput->dtForFrame = targetSecondsPerFrame;

                FILETIME newWriteTime = win32GetLastWriteTime(sourceGameCodeDLLFullPath);
                if (CompareFileTime(&newWriteTime, &game.lastWriteTime) != 0) {
                    win32UnloadGameCode(&game);
                    game = win32LoadGameCode(sourceGameCodeDLLFullPath, tempGameCodeDLLFullPath);
                }

                GameControllerInput *oldKeyboardController = &oldInput->controllers[0];
                GameControllerInput *newKeyboardController = &newInput->controllers[0];
                GameControllerInput zeroController = {};
                *newKeyboardController = zeroController;
                for (int buttonIndex = 0; buttonIndex < ArrayCount(oldKeyboardController->buttons); buttonIndex++) {
                    newKeyboardController->buttons[buttonIndex].endedDown = oldKeyboardController->buttons[buttonIndex].endedDown;
                }

                win32ProcessPendingMessages(&win32State, newKeyboardController);

                POINT mousePos;
                GetCursorPos(&mousePos);
                ScreenToClient(window, &mousePos);
                newInput->mouseX = mousePos.x;
                newInput->mouseY = mousePos.y;
                newInput->mouseZ = 0; // mouse wheel
                win32ProcessKeyboardKeypress(&newInput->mouseButtons[0], GetKeyState(VK_LBUTTON) & (1 << 15));
                win32ProcessKeyboardKeypress(&newInput->mouseButtons[1], GetKeyState(VK_MBUTTON) & (1 << 15));
                win32ProcessKeyboardKeypress(&newInput->mouseButtons[2], GetKeyState(VK_RBUTTON) & (1 << 15));
                win32ProcessKeyboardKeypress(&newInput->mouseButtons[3], GetKeyState(VK_XBUTTON1) & (1 << 15));
                win32ProcessKeyboardKeypress(&newInput->mouseButtons[4], GetKeyState(VK_XBUTTON2) & (1 << 15));

                XINPUT_STATE controllerState;
                int maxControllerCount = XUSER_MAX_COUNT;
                if (maxControllerCount > ArrayCount(newInput->controllers) - 1) {
                    maxControllerCount = ArrayCount(newInput->controllers) - 1;
                }

                for (DWORD controllerIndex = 0; controllerIndex < XUSER_MAX_COUNT; controllerIndex++) {
                    if (XInputGetState(controllerIndex, &controllerState) == ERROR_SUCCESS) {
                        DWORD ourControllerIndex = controllerIndex + 1;
                        GameControllerInput *oldController = &oldInput->controllers[ourControllerIndex];
                        GameControllerInput *newController = &newInput->controllers[ourControllerIndex];

                        newController->isAnalog = true;

                        XINPUT_GAMEPAD *pad = &controllerState.Gamepad;

                        int16 stickX = pad->sThumbLX;
                        int16 stickY = pad->sThumbLY;
                        real32 stickYNorm = 0;
                        real32 stickXNorm = 0;
                        if (stickY != 0 && abs(stickY) > XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE) {
                            stickYNorm = stickY / 32767.0f;
                        }
                        if (stickX != 0 && abs(stickX) > XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE) {
                            stickXNorm = stickX / 32767.0f;
                        }

                        newController->stickAvgX = stickXNorm;
                        newController->stickAvgY = stickYNorm;
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_UP) {
                            newController->stickAvgY = 1.0f;
                        }
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_DOWN) {
                            newController->stickAvgY = -1.0f;
                        }
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_LEFT) {
                            newController->stickAvgX = -1.0f;
                        }
                        if (pad->wButtons & XINPUT_GAMEPAD_DPAD_RIGHT) {
                            newController->stickAvgX = 1.0f;
                        }

                        //bool start = (pad->wButtons & XINPUT_GAMEPAD_START);
                        //bool back = (pad->wButtons & XINPUT_GAMEPAD_BACK);

                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnDown, &newController->btnDown, XINPUT_GAMEPAD_A);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnRight, &newController->btnRight, XINPUT_GAMEPAD_B);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnLeft, &newController->btnLeft, XINPUT_GAMEPAD_X);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->btnUp, &newController->btnUp, XINPUT_GAMEPAD_Y);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->leftShoulder, &newController->leftShoulder, XINPUT_GAMEPAD_LEFT_SHOULDER);
                        win32ProcessXInputDigitalButton(pad->wButtons, &oldController->rightShoulder, &newController->rightShoulder, XINPUT_GAMEPAD_RIGHT_SHOULDER);

                        real32 threshold = 0.5f;
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgY < -threshold), &oldController->stickDown, &newController->stickDown, 1);
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgX > threshold), &oldController->stickRight, &newController->stickRight, 1);
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgX < -threshold), &oldController->stickLeft, &newController->stickLeft, 1);
                        win32ProcessXInputDigitalButton((DWORD)(newController->stickAvgY > threshold), &oldController->stickUp, &newController->stickUp, 1);
                    } else {
                        // controller not available
                    }
                }

                ThreadContext threadCtx = {};

                GameOffscreenBuffer videoBuffer = {};
                videoBuffer.memory = globalBackBuffer.memory;
                videoBuffer.width = globalBackBuffer.width;
                videoBuffer.height = globalBackBuffer.height;
                videoBuffer.pitch = globalBackBuffer.pitch;
                videoBuffer.bytesPerPixel = globalBackBuffer.bytesPerPixel;

                if (win32State.inputRecordingIndex) {
                    win32RecordInput(&win32State, newInput);
                }
                if (win32State.inputPlayingIndex) {
                    win32PlaybackInput(&win32State, newInput);
                }

                if (game.updateAndRender) {
                    game.updateAndRender(&threadCtx, &gameMemory, &videoBuffer, newInput);
                }

                LARGE_INTEGER audioWallClock = win32GetWallClock();
                real32 fromBeginToAudioSeconds = win32GetSecondsElapsed(flipWallClock, audioWallClock);

                DWORD playCursor = 0;
                DWORD writeCursor = 0;
                if (SUCCEEDED(globalSecondaryBuffer->GetCurrentPosition(&playCursor, &writeCursor))) {
                    if (!soundIsValid) {
                        soundOutput.runningSampleIndex = writeCursor / soundOutput.bytesPerSample;
                        soundIsValid = true;
                    }
                    DWORD byteToLock = (soundOutput.runningSampleIndex*soundOutput.bytesPerSample) % soundOutput.secondaryBufferSize;
                    DWORD expectedSoundBytesPerFrame = (DWORD)((soundOutput.samplesPerSecond * soundOutput.bytesPerSample) / gameUpdateHz);
                    real32 secondsLeftUntilFlip = targetSecondsPerFrame - fromBeginToAudioSeconds;
                    DWORD expectedBytesUntilFlip = (DWORD)((secondsLeftUntilFlip/targetSecondsPerFrame) * (real32)expectedSoundBytesPerFrame);
                    DWORD expectedFrameBoundaryByte = playCursor + expectedBytesUntilFlip;

                    DWORD safeWriteCursor = writeCursor;
                    if (safeWriteCursor < playCursor) {
                        safeWriteCursor += soundOutput.secondaryBufferSize;
                    }
                    Assert(safeWriteCursor >= playCursor);
                    safeWriteCursor += soundOutput.safetyBytes;

                    bool audioIsLowLatency = safeWriteCursor < expectedFrameBoundaryByte;

                    DWORD targetCursor = 0;
                    if (audioIsLowLatency) {
                        targetCursor = expectedFrameBoundaryByte + expectedSoundBytesPerFrame;
                    } else {
                        targetCursor = writeCursor + expectedSoundBytesPerFrame + soundOutput.safetyBytes;
                    }
                    targetCursor %= soundOutput.secondaryBufferSize;

                    DWORD bytesToWrite = 0;
                    if (byteToLock > targetCursor) {
                        bytesToWrite = soundOutput.secondaryBufferSize - byteToLock + targetCursor;
                    } else {
                        bytesToWrite = targetCursor - byteToLock;
                    }

                    GameSoundOutputBuffer soundBuffer = {};
                    soundBuffer.samplesPerSecond = soundOutput.samplesPerSecond;
                    soundBuffer.sampleCount = bytesToWrite / soundOutput.bytesPerSample;
                    soundBuffer.samples = samples;
                    game.getSoundSamples(&threadCtx, &gameMemory, &soundBuffer);
                    win32FillSoundBuffer(&soundOutput, byteToLock, bytesToWrite, &soundBuffer);
                } else {
                    soundIsValid = false;
                }


                real32 secondsElapsedForFrame = win32GetSecondsElapsed(lastWorkCounter, win32GetWallClock());
                if (secondsElapsedForFrame < targetSecondsPerFrame) {
                    if (sleepIsGranular) {
                        DWORD sleepMs = (DWORD)(1000.0f * (targetSecondsPerFrame - secondsElapsedForFrame));
                        if (sleepMs > 0) {
                            Sleep(sleepMs);
                        }
                    }
                    while (secondsElapsedForFrame < targetSecondsPerFrame) {
                        secondsElapsedForFrame = win32GetSecondsElapsed(lastWorkCounter, win32GetWallClock());
                    }
                } else {
                    // logging, MISSED FRAME!
                }

                lastWorkCounter = win32GetWallClock();

                win32DrawBufferInWindow(&globalBackBuffer, window);
                flipWallClock = win32GetWallClock();

#if 0
                real64 msElapsed = (real32)(1000.0f*secondsElapsedForFrame) / globalPerfCountFrequency;
                real64 fps = (real32)(1000.0f*globalPerfCountFrequency/(real32)secondsElapsedForFrame)/1000.0f;
                int64 endCycleCount = __rdtsc();
                int64 cyclesElapsed = endCycleCount - lastCycleCount;
                lastCycleCount = endCycleCount;
#endif

                GameInput *temp = newInput;
                newInput = oldInput; 
                oldInput = temp;
            }
        } else {
            // failed
        }
    } else {
        // failed
    }

    return 0;
}

